using System;
using System.Numerics;
using System.Threading;
using System.Threading.Tasks;

namespace FastRng.Distributions;

public abstract class Distribution<TNum> : IDistribution<TNum> where TNum : IFloatingPointIeee754<TNum>
{
    private readonly ShapeFitter<TNum> fitter;

    protected Distribution(IRandom<TNum> rng)
    {
        if (rng == null)
            throw new ArgumentNullException(nameof(rng), "An IRandom<TNum> implementation is needed.");
        
        this.fitter = new ShapeFitter<TNum>(this.ShapeFunction, rng, 100);
    }

    private protected abstract TNum ShapeFunction(TNum x);
        
    public async ValueTask<TNum> GetDistributedValue(CancellationToken token = default) => await this.fitter.NextNumber(token);
        
    public async ValueTask<uint> NextNumber(uint rangeStart, uint rangeEnd, CancellationToken cancel = default)
    {
        // Swap the values if the range start is greater than the range end:
        if (rangeStart > rangeEnd)
            (rangeStart, rangeEnd) = (rangeEnd, rangeStart);

        var range = rangeEnd - rangeStart;
        var distributedValue = await this.GetDistributedValue(cancel);
        return (uint) ((float.CreateChecked(distributedValue) * range) + rangeStart);
    }

    public async ValueTask<ulong> NextNumber(ulong rangeStart, ulong rangeEnd, CancellationToken cancel = default)
    {
        // Swap the values if the range start is greater than the range end:
        if (rangeStart > rangeEnd)
            (rangeStart, rangeEnd) = (rangeEnd, rangeStart);

        var range = rangeEnd - rangeStart;
        var distributedValue = await this.GetDistributedValue(cancel);
        return (ulong) ((double.CreateChecked(distributedValue) * range) + rangeStart);
    }

    public async ValueTask<TNum> NextNumber(TNum rangeStart, TNum rangeEnd, CancellationToken cancel = default)
    {
        // Swap the values if the range start is greater than the range end:
        if (rangeStart > rangeEnd)
            (rangeStart, rangeEnd) = (rangeEnd, rangeStart);

        var range = rangeEnd - rangeStart;
        var distributedValue = await this.GetDistributedValue(cancel);
        return (distributedValue * range) + rangeStart;
    }

    public async ValueTask<TNum> NextNumber(CancellationToken cancel = default) => await this.NextNumber(TNum.Zero, TNum.One, cancel);
        
    public async ValueTask<bool> HasDecisionBeenMade(TNum above, TNum below, CancellationToken cancel = default)
    {
        var number = await this.NextNumber(cancel);
        return number > above && number < below;
    }
}